#!/usr/bin/python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import os.path
import subprocess
import gobject

from tacix.util import *

class TXMount:
    def __init__ (self, sessionid):
        self.txsession = get_session(sessionid)
        self.txpath = TXPath(sessionid)
        self.logger = self.txpath.get_logger("txmount")

        self.signal = None
        self.mounts = []
        self.counters = {}

    def watch_agent_state(self, state):
        self.logger.info("Mount watcher state is %s" % state)
        if state == 'running':
            self.mount_pending()
        elif state == 'terminating' or state == 'suspending':
            self.remove_mounted()
        elif state == 'terminated' or state == 'suspended':
            self.remove_mounted(forced=true)
            self.txsession.ClearMounts()

    def remove_mounted(self, force=False):
        self.logger.debug('Removing all mounted %s' % self.mounts)
        for mount_name in self.mounts:
            self.remove_mount(mount_name, force=force)

    def mount_pending(self):
        self.logger.debug('Mounting all pending')
        pending_mounts = self.txsession.GetPendingMounts()
        for mount_name in pending_mounts:
            gobject.timeout_add(1000, self.add_mount, mount_name)

    def setup_signal(self):
        self.logger.debug('Adding watcher for TXMount')
        self.signal = self.txsession.connect_to_signal("StateChanged", self.watch_agent_state, dbus_interface=SESSION_INTERFACE)
        self.mountsolicited_signal = self.txsession.connect_to_signal("MountSolicited", self.watch_mount_solicited, dbus_interface=SESSION_INTERFACE)
        if self.txsession.GetState() == "running":
            self.mount_pending()

    def watch_mount_solicited(self, mount_name):
        self.logger.debug('Mount %s solicited' % mount_name)
        if self.txsession.GetState() == 'running':
            self.add_mount(mount_name)

    def add_mount(self, mount_name):
        self.logger.debug('Adding mount %s' % mount_name)
        options = self.txsession.GetMount(mount_name)
        self.logger.debug('Mount parameters %s' % options)
        if not os.path.exists("%(dir)s" % options):
            os.makedirs("%(dir)s" % options)
        cmd = ["/sbin/mount.cifs",
                "//%(computername)s/%(share)s" % options,
                "%(dir)s" % options,
                "-n", #Do not write to mtab
                "-o",
                "username=\"%(username)s\",ip=127.0.0.1,port=%(port)s" % options]
        process = subprocess.Popen(cmd, close_fds=True, universal_newlines=True, stdin=subprocess.PIPE, stdout=None, stderr=None)
        process.communicate(options['password'] + "\n")
        ret = process.wait()
        if ret != 0:
            self.logger.info('Failed to mount %s' % mount_name)
            if mount_name not in self.counters:
                self.counters[mount_name] = 30
            counter = self.counters[mount_name] - 1
            self.counters[mount_name] = counter
            if counter == 0:
                self.counters.remove(mount_name)
                self.txsession.Notify('TaciX mount support', 'Failed to mount share %(share)s at %(dir)s.' % (options))
                return False
            return True

        self.logger.debug('Mounted %s' % mount_name)
        self.mounts.append(options['dir'])
        self.txsession.SetMounted(mount_name)
        self.txsession.Notify('TaciX mount support', 'Share %(share)s mounted at %(dir)s.' % (options))
        if mount_name in self.counters:
            self.counters.remove(mount_name)
        return False

    def remove_mount(self, mount_name, force=False):
        self.logger.debug('Removing mount %s' % mount_name)
        cmd = ["/sbin/umount.cifs", mount_name]
        if force:
            cmd.append("-f")
        self.logger.debug('%s' % cmd)
        ret = subprocess.call(cmd, stdin=None, stdout=None, stderr=None)
        self.logger.debug("Umount returned %s" % ret)
        if ret == 0:
            self.mounts.remove(mount_name)
