#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import sys
import dbus
import gobject
import traceback

import dbus.mainloop.glib
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
import dbus.glib

from tacix.server.nx.node import NXNode
from tacix.freenx.FreeNXNodeProtocol import FreeNXNodeProtocol
from tacix.freenx.FreeNXShell import *
from tacix.freenx.FreeNXServerProtocol import FreeNXServerProtocol
from tacix.util import *

class FreeNXNode(NXNode, FreeNXNodeProtocol, FreeNXShell, FreeNXServerProtocol):
    def __init__( self, logger=None, txpath=None ):
        FreeNXShell.__init__(self, logger=logger)
        FreeNXNodeProtocol.__init__(self)
        NXNode.__init__(self, logger=logger)

        self.tacix_mode = False
        self.waiting_restore = False

        self.txpath = txpath
        if self.logger is None and self.txpath is not None:
            self.txpath.get_logger('freenxnodeslave')

    # Watchers
    def watch_agent_state_resume(self, state):
        if state == 'running':
            self.say_session_resumed()
            sys.stdout.write('NX> 1005 Session status: %s\n' % state)
        elif state == 'failed':
            self.say_session_resume_failed()
            self.say_bye()
        else:
            self.watch_agent_state(state)

    def watch_agent_state(self, state):
        # On session supended remove signals
        if self.logger is not None:
            self.logger.info("Nxagent state is %s" % \
                state)

        if self.waiting_restore and state == 'suspended':
            self.waiting_restore = False
            return
        if self.tacix_mode:
            sys.stdout.write('NX> 1005 Session status: %s\n' % state)
        else:
            if state == 'suspended':
                sys.stdout.write('NX> 1005 Session status: %s\n' % state)
            elif state != 'waiting_connection' and state != 'denied':
                sys.stdout.write('NX> 1009 Session status: %s\n' % state)

        if state == 'waiting_connection':
            self.say_connected()
            self.say_waiting()

        if state in ('suspended', 'failed', 'terminated', 'denied'):
            self.logger.debug("Exiting NXNode")
            self.remove_signals()
            self.loop.quit()

            if state == 'suspended':
                self.say_bye()
                sys.stdout.flush()
            elif state == 'denied':
                self.say_abort('Permission denied')
                self.say_bye()
            elif state == 'terminated':
                self.say_closed()
                self.say_bye()
            elif state == 'failed':
                self.say_error()
                self.say_bye()

        sys.stdout.flush()

    # Actions
    def list_session(self, options):
        try:
            if options['type'] == 'shadow':
                self.say_shadowsession_header(options)
                sessions = self.provider.GetShadowableSessions()
                for session_key in sessions.keys():
                    self.say_listsessions_sessions(sessions[session_key])

            else:
                self.say_listsessions_header(options)
                sessions = self.provider.GetResumableSessions (options)
                for session_opts in sessions.values():
                    self.logger.debug(session_opts)
                    self.say_listsessions_sessions(session_opts)
        except dbus.DBusException, e:
            self.logger.debug( 'DBusError: %s' % e)
            raise FreeNXException( e._dbus_error_name )

        self.say_listsessions_footer(options)


    # Helpers
    def _set_tacix_mode_handler(self, line):
        self.tacix_mode = True
        if self.logger is not None:
            self.logger.debug('Starting tacix mode')
        return True

    def _startsession_handler(self, line):
        opts = self.opt_read(line)
        self.start_session(opts, 'nx')
        return True

    def _listsession_handler(self, line):
        opts = self.opt_read(line)
        if self.logger is not None:
            self.logger.debug('Listing sessions %s' % line)

        self.list_session(opts)
        return True

    def _restoresession_handler(self, line):
        self.waiting_restore = True
        opts = self.opt_read(line)
        self.restore_session(opts)
        return True

    def _attachsession_handler(self, line):
        print("NX> 726 Asking user for authorization to attach to session")
        sys.stdout.flush()
        opts = self.opt_read(line)
        self.attach_session(opts)
        self.wait_session()
        return True

    def _addmount_handler(self, line):
        # addmount  --port="4100" --username="user" --password="***" --share="SharedDocs" --computername="CASA" --session_id="A9AE3F149BCD5C39C435EC0C1A64AD20" --dir="/home/mshima/SharedDocs"
        opts = self.opt_read(line)
        self.add_mount(opts)
        return True

    def _addprinter_handler(self, line):
        #['addprinter  --type="ipp" --username="user" --port="+3000" --session_id="4022DFDCEAAA1829A1AFA33124D881BC" --printer="IBM-Compatible_Dot_Matrix" --password="***" --model="NULL"\n']
        opts = self.opt_read(line)
        self.add_printer(opts)
        return True

    def _terminate_handler(self, line):
        opts = self.opt_read(line)
        # Terminate send sessionid instead of id so copy it
        opts['id'] = opts['sessionid']
        self.terminate_session(opts)
        return True

    def _check_handler(self, line):
        self.check()
        sys.stdout.flush()
        return True

    def _quit_handler(self, line):
        return False

    def _bye_handler(self, line):
        self.say_server_bye(line)
        self.logger.debug('Waiting session to finish')
        self.wait_session()
        return True

    def slave_loop(self):
        '''
        Init the FreeNX Node in slave mode
        '''
        self.say_hello()
        self.say_slave_started()
        sys.stdout.flush()
        gobject.io_add_watch(sys.stdin, gobject.IO_IN | gobject.IO_HUP, self.read_line)
        self.wait_session()

    def read_line(self, source, condition):
        if condition == gobject.IO_HUP:
            self.quit()
            return False
        self.set_non_blocking(sys.stdin.fileno())
        ret = False
        try:
            ret = self.receive_stdin()
        except Exception, e:
            self.say_abort(e)
            trace = traceback.format_exc()
            self.logger.error('Exiting on Exception.')

            for line in trace.split('\n'):
                self.logger.error('%s' % line)
            self.quit()
        self.revert_blocking(sys.stdin.fileno())
        return ret


