#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import sys
import select
import fcntl
import os
import re

class FreeNXException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)
    def __repr__(self):
        return repr(self.value)

class FreeNXShell():
    def __init__( self, logger=None ):
        self.logger = logger
        self.flags = None

    def set_non_blocking(self, fd):
        """Make a fd non-blocking."""
        self.flags = fcntl.fcntl(fd, fcntl.F_GETFL)
        flags = self.flags | os.O_NONBLOCK
        fcntl.fcntl(fd, fcntl.F_SETFL, flags) 

    def revert_blocking(self, fd):
        """Make a fd non-blocking."""
        fcntl.fcntl(fd, fcntl.F_SETFL, self.flags) 

    def opt_read(self, opt_str):
        """Parse options into a dict."""
        options = {}
        commands = opt_str.split(" --")
        if len(commands) <= 1:
            return options

        for command in commands[1:]:
            split = command.split('=')
            key = split[0]
            value = split[1].replace('"', '')
            self.logger.debug('Param matched: %r=%r' % (key, value))
            if key is None or value is None:
                continue
            options[key] = value
        return options

    def _receive_stdin(self, fn_handler, allowed_cmds=None):
        read = sys.stdin.read()
        return self._receive(read, fn_handler, allowed_cmds=allowed_cmds)

    def _receive(self, read, fn_handler, allowed_cmds=None):
        if len(read) == 0:
            if self.logger is not None:
                self.logger.debug("EOF received")
            return False

        lines = read.splitlines(True)
        if self.logger is not None:
            self.logger.debug(lines)

        ret = True
        for line in lines:
            if allowed_cmds is None:
                ret = fn_handler(line)
            else:
                cmd = line.split()[0].lower()
                if cmd in allowed_cmds:
                    return self.handle_line(line)
        return ret

    def receive_stdin(self):
        return self._receive_stdin(self.handle_line)

    def _loop(self, handlers, message='waiting'):
        while True:
            if self.logger is not None:
                self.logger.debug(message)

            r, w, e = select.select(handlers.keys(), [], [])
            ret = True
            for key in handlers.keys():
                if key in r:
                    ret = handlers[key]()
                    if not ret:
                        break
            if not ret:
                break

    def stdin_loop(self):
        self._loop({sys.stdin:self.receive_stdin})

    def handle_line(self, line, prefix=None):
        if self.logger is not None:
            self.logger.debug('stdin:' + line)
        if line is None: 
            return True
        line = line.strip()
        if line == '':
            return True

        cmd = line.split()[0].lower()
        handler_name = '_%s_handler' % cmd
        if prefix is not None:
            handler_name = '_' + prefix + handler_name
        try:
            handler_method = getattr(self, handler_name)
            self.logger.debug('handler %s found' % handler_name)
            return handler_method(line)
        except AttributeError, e:
            self.logger.debug('Unhandled nx command %r (%s)' % (cmd, e))
            raise FreeNXException('%s' % e)

        return True

    def run( self ):
        self.set_non_blocking(sys.stdin.fileno())
        self.stdin_loop()
        self.revert_blocking(sys.stdin.fileno())

