#!/usr/bin/python
#
# Copyright 2007 Google Inc.
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: diamond@google.com (Stephen Shirley)
#         marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import re
import signal
import traceback
import gobject


state_lines = {
        'starting': re.compile(r'^Session: Starting session at '),
        'resuming': re.compile(r'^Session: Resuming session at '),
        'terminating': re.compile(r'^Session: (Terminat|Abort)ing session at '),
        'running': re.compile(r'^Session: Session (started|resumed) at '),
        'suspending': re.compile(r'^Session: Suspending session at '),
        'suspended': re.compile(r'^Session: Session suspended at '),
        'terminated': re.compile(r'^Session: Session (terminat|abort)ed at '),
        'failed': re.compile(r'^Session: Display failure detected at ')
    }
info_lines = {
        'smb_port': re.compile(r"^Info: Listening to SMB connections on port '(?P<port>\d+)'."),
        'watchdog_pid': re.compile(r"^Info: Watchdog running with pid '(?P<pid>\d+)'."),
        'kill_watchdog': re.compile(r"^Info: Waiting the watchdog process to complete."),
        'agent_pid': re.compile(r"^Info: Agent running with pid '(?P<pid>\d+)'."),
        'waiting_connection': re.compile(r"^Info: Waiting for connection from "),
        'general_error': re.compile(r"^Error: (?P<error>.*)$")
    }

class AgentMonitor:
    def __init__( self, nxsession ):
        self.txsession = nxsession.txsession
        self.nxagent = nxsession.nxagent
        self.txpath = nxsession.txpath
        self.nxsession = nxsession

        self.logger = self.txpath.get_logger("nxagentmonitor")

        self.session_log = open(self.txpath.get_filepath( "session" ), "w")

    def read_line(self, source, condition):
        if condition == gobject.IO_HUP:
            self.nxsession.change_state('terminated')
            return False
        try:
            line = self.nxagent.app_process.stdout.readline()
            if line is None: 
                return True
            self.session_log.write(line)
            line = line.strip() # Remove newline
            self.logger.info("Readline: condition = %s, str = %s" % (condition, line))

            if line == '':
                return True

            if line.startswith('Session: '):
                self.match_state(line)
                self.logger.debug("Matched state: %s\n" % line)
            elif line.startswith('Info: ') or line.startswith('Error: '):
                self.match_info(line)
                self.logger.debug("Matched info: %s\n" % line)

            self.session_log.flush()

        except Exception, e:
            self.logger.error("Got exception, cleaning up\n")
            trace = traceback.format_exc()
            for line in trace.split('\n'):
                self.logger.error('%s' % line)

            return True
        return True


    def run(self):
        gobject.io_add_watch(self.nxagent.app_process.stdout, gobject.IO_IN | gobject.IO_HUP, self.read_line)


    def match_state(self, line):
        """Try match the given line against a session state

        Return:
            True: the line matched
            False: the line didn't match
        """
        for state, rx in state_lines.iteritems():
            m = rx.search(line)
            if m:
                self.nxsession.change_state(state)
                return True
        return False


    def match_info(self, line):
        """Try match the given line against an info line regex
    
        Return:
            True: the line matched
            False: the line didn't match
        """
        for info, rx in info_lines.iteritems():
            m = rx.search(line)
            if m:
                self.handle_info(info, m)
                return True
        return False


    def handle_info(self, info, m):
        """Execute the required response to a given info line match
    
        Args:
            info: The id of the info line
            m: The match object from the info line's regex
        Return:
            None
        """

        value = ''
        if info == 'watchdog_pid':
            self.nxsession.watchdog_pid = m.group('pid')
            return
        if info == 'agent_pid':
            value = m.group('pid')

        elif info == 'smb_port':
            value = m.group('port')

        elif info == 'kill_watchdog':
            self.nxsession.change_state('waiting_watchdog')
            return
        elif info == 'general_error':
            value = m.group('error')
            self.logger.error("Agent error: %s" % value)

        elif info == 'waiting_connection':
            self.nxsession.change_state('waiting_connection')
            return
        else:
            # If none of the above handers match...
            self.logger.error("matched info %s, but failed to "
                    "find handler for it\n" % info)
            self.txsession.SendInfo(info)
            return

        self.logger.debug("Setting dbus txsession info: %s and value: %s" % (info, value))
        if value == '':
            self.txsession.SendInfo(info)
        else:
            self.txsession.SetProperty(info, value)
            self.nxagent.options[info] = value

