#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import sys
import subprocess
import dbus
import signal
import getpass
import gobject

from tacix.server.nx import NXException
from tacix.util import *
from tacix.constants import *

import dbus.mainloop.glib
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

class NXNode:
    def __init__( self, logger=None, user=None, loop=gobject.MainLoop() ):
        self.txsession = None
        self.logger = logger
        self.txpath = None
        self.state_signal = None
        self.info_signal = None

        self.loop = loop

        self.bus = dbus.SystemBus()
        self.manager_obj = self.bus.get_object (BUS_NAME, MANAGER_PATH)
        self.manager = dbus.Interface (self.manager_obj, MANAGER_INTERFACE)
        self.provider = dbus.Interface (self.bus.get_object (PROVIDER_BUS_NAME, PROVIDER_PATH), PROVIDER_INTERFACE)
        self.user = user
        if user is None:
            self.user = getpass.getuser()

    # Watchers
    # On session supended remove signals
    def watch_agent_info(self, info):
        '''
        This method should be overrided
        '''
        pass

    def watch_agent_state(self, state):
        '''
        This method should be overrided
        '''
        pass

    def __watch_agent_state(self, state):
        if state == 'ready' and self.txsession.GetProperty ('seat-type') == 'nx':
            self.txsession.Run ()
            self.__state_signal.remove ()
        elif state == 'ready':
            self.logger.debug ('Starting shadow')
            self.session_worker = dbus.Interface (self.bus.get_object (self.txsession.GetSessionWorker(), '/Session'), SESSION_WORKER_INTERFACE)

            self.session_worker.RunShadowForCKSession (self.txsession.GetProperty ('shadowed_id'))
            self.logger.debug ('Running shadow')
            self.txsession.Run ()
            self.__state_signal.remove ()
            self.logger.debug ('shadow running')
        return

    def watch_agent_state_resume(self, state):
        '''
        This method should be overrided
        '''
        pass

    # Helpers
    def get_property(self, property):
        return self.txsession.GetProperty(property)

    def add_state_signal(self, func):
        self.state_signal = self.txsession.connect_to_signal("StateChanged", func, dbus_interface=SESSION_INTERFACE)

    def add_info_signal(self, func):
        self.info_signal = self.txsession.connect_to_signal("InfoReceived", func, dbus_interface=SESSION_INTERFACE)

    def add_signals(self):
        self.add_state_signal(self.watch_agent_state)
        self.add_info_signal(self.watch_agent_info)
        self.logger.debug('Signals added')

    def remove_signals(self):
        if self.state_signal is not None:
            self.state_signal.remove()
        if self.info_signal is not None:
            self.info_signal.remove()
        self.logger.debug('Signals removed')


    # Actions functions
    def restore_session(self, options):
        self.get_session(options['id'])
        self.txsession.SetProperties( options )

        self.add_state_signal(self.watch_agent_state_resume)
        self.add_info_signal(self.watch_agent_info)
        self.txsession.Resume()

    def add_printer(self, options):
        self.txsession.AddPrinter(options)

    def add_mount(self, options):
        self.txsession.AddMount(options)

    def attach_session(self, options):
        if 'id' in options:
            options['shadowed_id'] = options['id']
            del options['id']
        if 'display' in options:
            options['shadowdisplay'] = options['display']
            del options['display']
        self.start_session(options, 'nx-shadow')

    def suspend_session(self, options):
        self.get_session(options['id'])
        self.txsession.Suspend()
        #self.send_signal(signal.SIGHUP)

    def terminate_session(self, options):
        self.get_session(options['id'])
        self.txsession.Terminate()
        #self.send_signal(signal.SIGTERM)

    def get_session(self, session_id):
        session_path = self.manager.GetSession (session_id)
        session_obj = self.bus.get_object (BUS_NAME, session_path)
        self.txsession = dbus.Interface (session_obj, SESSION_INTERFACE)

    def start_session(self, options, session_type):
        '''Init the NX Node'''
        # Create TXManager session
        session_path = self.manager.CreateSession (self.user, session_type)
        session_obj = self.bus.get_object (BUS_NAME, session_path)
        self.txsession = dbus.Interface (session_obj, SESSION_INTERFACE)
        self.add_signals()
        self.__state_signal = self.txsession.connect_to_signal("StateChanged", self.__watch_agent_state, dbus_interface=SESSION_INTERFACE)
 
        self.txsession.SetProperties( options )

        session_id = self.txsession.GetProperty('id')
        self.txpath = TXPath( session_id, create=True )
        if self.logger is None:
            self.logger = self.txpath.get_logger("nxnode")

    def wait_session(self):
        '''
        Wait for the session to quit
        '''
        self.logger.debug("Waiting for jobs to complete")
        self.loop.run()

    def quit(self):
        self.loop.quit()


