#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import dbus
import dbus.service
import os, os.path
import subprocess
import pwd
import random

from tacix import xauth
from tacix.util import *

try:
    from hashlib import md5
except ImportError:
    from md5 import md5

class CKException(dbus.DBusException):
    _dbus_error_name = 'org.tacix.session.CKException'

class TaciXSessionException(dbus.DBusException):
    _dbus_error_name = 'org.tacix.session.TaciXSessionException'

def gen_cookie():
    return md5(str(random.random())).hexdigest()

class Session( dbus.service.Object ):
    def __init__( self, sender, session_type, bus, manager, ck_manager, dbus_info, logger, effective_user=None, provider=None, ck_session=None):
        '''
        effective_user is the user that will run the agent
        '''
        self.logger = logger
        self.seat_type = session_type

        # Getting the system dbus
        #self.bus = dbus.SystemBus( )
        self.bus = bus

        self.ck_manager = ck_manager
        #self.ck_manager = dbus.Interface (
        #        self.bus.get_object ('org.freedesktop.ConsoleKit', '/org/freedesktop/ConsoleKit/Manager'),
        #        'org.freedesktop.ConsoleKit.Manager')

        # Attributes
        self.manager = manager
        self.dbus_info = dbus_info
        self.provider = provider

        self.authority_file = None
        self.display = None
        self.state = None

        self.ck_session = ck_session
        self.ck_cookie = None
        self.ck_session_id = None

        self.session_worker = None
        self.session_worker_conn = None


        self.guest_session = False
        if type == 'nx-guest':
            self.guest_session = True
        #elif type == 'nx-shadow':

        if self.guest_session:
            guest_name = self.create_guest_user()
            if guest_name is not None:
                effective_user = guest_name

        if effective_user is None:
            return
        self.effective_pwd = pwd.getpwnam(effective_user)

        self.display = provider.FindNextAvailableDisplay()
        self.ck_cookie = self._create_ck_session(self.effective_pwd.pw_uid, ":" + self.display, 'remote_host')
        self.ck_session_id = self.ck_session.GetId().rsplit('/', 2)[-1]
        self.cookie = gen_cookie()

        # Properties
        if self.seat_type == 'xrdp-vnc':
            pass
        else:
            if sender is not None:
                self.logger.debug( "Watch for conn: %s" % sender )
                self.node_signal = self.bus.watch_name_owner(sender, self.connection_node_closed_signal)

        self.properties = provider.GetDefaultOptions()
        self.properties.update(self.provider.CalculateOptions(str(self.effective_pwd.pw_uid), self.display, self.ck_session_id))

        self.object_path = dbus.ObjectPath(SESSION_PATH + self.ck_session_id)
        dbus.service.Object.__init__(self, bus, self.object_path)

        self.authority_file = xauth.create_xauthority_file_for_user(effective_user, self.cookie,
                ":" + self.display)
        if self.authority_file is None:
            self.logger.debug("Could not create the authority file")

        self.set_property('authority_file', self.authority_file)
        self.set_property('seat-type', session_type)
        self.session_process = run_as_user(self.effective_pwd, self.ck_session_id, session_type)

        self.watch_connection = False


    def __run_async_rec(self, success):
        if not success:
            self._close()

    def __async_error_handler(self, error):
        self._close()

    def get_object_path(self):
        return self.object_path

    def get_property(self, property):
        if property not in self.properties:
            return None
        return self.properties[property]

    def get_property_complete(self, property):
        if property == 'cookie':
            return self.cookie
        return self.get_property(property)

    def get_properties(self):
        ret = {}
        ret.update(self.properties)
        return ret

    def get_properties_complete(self):
        ret = {'cookie':self.cookie}
        ret.update(self.properties)
        return ret

    def set_property(self, property, value):
        val = self.provider.IsAllowed(property, value)
        self.logger.debug(val)
        if val is not None:
            self.properties.update(val)

    def run_dialog(self, authority_file, display):
        ret = xauth.run_program("/usr/lib/tacix/tacix-dialog", authority_file, display)
        self.logger.debug('tacix-dialog returned %d' % ret)
        if os.path.lexists(authority_file):
            os.remove(authority_file)
        return ret

    def ask_console_authority(self, sender):
        display = ":" + self.get_property('shadowdisplay')

        ck_sessions = self.ck_manager.GetSessions()
        local_ck_session = None
        for ck_op_session in ck_sessions:
            ck_session_obj = self.bus.get_object ('org.freedesktop.ConsoleKit', ck_op_session)
            ck_session = dbus.Interface (ck_session_obj, 'org.freedesktop.ConsoleKit.Session')
            ck_display = ck_session.GetX11Display()
            self.logger.debug('Comparing %s and %s' % (display, ck_display))
            if display == ck_display:
                self.logger.debug('ConsoleKit session found')
                local_ck_session = ck_session
                break

        pwd_user = pwd.getpwuid(local_ck_session.GetUser())
        console_authority_file = '%s/.Xauthority' % pwd_user.pw_dir

        if self.bus.get_unix_user(sender) == local_ck_session.GetUnixUser():
            return xauth.merge_cookie_from_xauthority(console_authority_file, self.authority_file, display)

        if local_ck_session == None:
            self.logger.debug('ConsoleKit session not found')
            return False

        tmp_authority_file = "/tmp/.nx-authority-%s" % str(random.random())
        self.logger.debug('Asking to merge display %s on %s into %s' %\
                (display, console_authority_file, tmp_authority_file))
        if not xauth.merge_cookie_from_xauthority(console_authority_file, tmp_authority_file, display):
            self.logger.debug('Failed to merge from console authority')
            self.logger.debug('console = %s tmp = %s display = %s ' %\
                    (console_authority_file, tmp_authority_file, display))
            return False

        if self.run_dialog(tmp_authority_file, display) == 1:
            return xauth.merge_cookie_from_xauthority(console_authority_file, self.authority_file, display)
        return False

    def privileged_run(self, sender, action):
        # Run action if the user is the sender
        if self.bus.get_unix_user(sender) == self.effective_pwd.pw_uid:
            action()
            return True
        return False

    def connection_node_closed_signal(self, sender):
        if not self.watch_connection:
            return
        if sender == '':
            self.logger.debug("Connection with node closed")
            self.node_signal.remove()
            if self.state != 'suspended' and self.state != 'suspending':
                self.CloseSession()

    def connection_session_closed_signal(self, sender):
        if not self.watch_connection:
            return
        if sender == '':
            self.logger.debug("Connection closed, session %s exited " % self.ck_session_id)
            self.CloseSession()

    def create_guest_user(self):
        process = subprocess.Popen("/usr/lib/tacix/guest-session-setup.sh", stdout=subprocess.PIPE, stderr=None, stdin=None)
        user = None
        while True:
            line = process.stdout.readline()
            if len(line) == 0:
                break
            user = line
        ret = process.wait()
        if ret == 0 and user is not None:
            user = user.strip()
            self.guest_user = user
            return user
        return None

    def cleanup_guest_user(self):
        ret = process = subprocess.call(["/usr/lib/tacix/guest-session-cleanup.sh", self.guest_user])
        if ret == 0:
            return True
        return False

    # Signals
    @dbus.service.signal( SESSION_INTERFACE )
    def StateChanged (self, state):
        self.state = state
        self.set_property('state', self.state)
        self.logger.debug( "StateChanged emitted, state = %s" % state )

    @dbus.service.signal( SESSION_INTERFACE )
    def PrinterAdded (self, printer_name):
        self.logger.debug( "Printer %s added" % printer_name )

    @dbus.service.signal( SESSION_INTERFACE )
    def MountAdded (self, mount_name):
        self.logger.debug( "Mount %s added" % mount_name )

    @dbus.service.signal( SESSION_INTERFACE )
    def InfoReceived (self, property):
        self.logger.debug( "InfoReceived emitted, property = %s" % property )

    @dbus.service.signal( SESSION_INTERFACE )
    def SessionClosed (self, message):
        self.logger.debug( "SessionClosed emitted" )

    @dbus.service.signal( SESSION_INTERFACE )
    def SessionNotify (self, title, message):
        self.logger.debug( "SessionNotify emitted" )

    # Methods 
    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def GetStatus(self, sender, conn):
        if self.session_worker is None:
            raise TaciXSessionException('SessionWorker not running')
        return self.session_worker.GetStatus()

    ### Authorization Methods ###
    def GetX11AuthorityFile(self):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.GetX11AuthorityFile
        '''
        return self.authority_file

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def GetCookie (self, sender, conn):
        return self.cookie

    def AddUserAuthorization(self, username):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.AddUserAuthorization\
          string:"value"
        '''
        return xauth.create_xauthority_file_for_user(username, self.cookie,
                ":" + self.display)

    def RemoveUserAuthorization(self, username):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.RemoveUserAuthorization \
          string:"value"
        '''
        return

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='s', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def RegisterSessionWorker(self, cookie, sender, conn):
        if self.session_worker_conn is not None or self.session_worker is not None:
            raise TaciXSessionException('Session Worker already registered')
        self.session_worker_conn = sender
        self.session_worker = dbus.Interface (self.bus.get_object (sender, '/Session'), 'org.tacix.session.worker')
        self.bus.watch_name_owner(sender, self.connection_session_closed_signal)
        self.session_worker.connect_to_signal("StateChanged", self.StateChanged, dbus_interface='org.tacix.session.worker')
        self.logger.debug('Session Worker registered ' + sender)
        return True

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def GetSessionWorker (self, sender, conn):
        if self.session_worker_conn is None:
            raise TaciXSessionException('SessionWorker not running')
        return self.session_worker_conn

    #@dbus.service.method( dbus_interface=SESSION_INTERFACE,
    #                   in_signature='s', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    #def RegisterSessionAgent(self, cookie, sender, conn):
    #
    #    return True

    def RemoveAuthorization(self, file):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.RemoveUserAuthorization \
          string:"value"
        '''
        xauth.remove_xauthority_file(file)
        return

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='ss', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def SetProperty( self, property, value, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.SetProperty \
          string:"value" \
          string:"value"
        '''
        self.logger.debug( "Setting property %s with value %s" % (property, value) )
        self.set_property(property, value)

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='a{ss}', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def SetProperties( self, properties, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.SetProperties \
          dict:string:string:"key0","value0","key1","value1"
        '''
        self.logger.debug( "Setting properties %s" % properties )
        for prop in properties:
            self.SetProperty(prop, properties[prop], sender, conn)

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='s', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def GetProperty( self, property, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.GetProperty \
          string:"value"
        '''
        if self.bus.get_unix_user(sender) == self.effective_pwd.pw_uid:
            return self.get_property_complete(property)
        return self.get_property(property)

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='a{ss}', sender_keyword='sender', connection_keyword='conn' )
    def GetProperties( self, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.GetProperties
        '''
        if self.bus.get_unix_user(sender) == self.effective_pwd.pw_uid:
            properties = self.get_properties_complete()
        else:
            properties = self.get_properties()
        self.logger.debug( "Getting properties %s" % properties )
        return properties

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='s', sender_keyword='sender', connection_keyword='conn' )
    def GetState( self, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.GetState
        '''
        return self.state

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def Close( self, sender=None, conn=None ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.Close
        '''
        self._close()

    def _close(self):
        if self.ck_session is not None:
            self.ck_manager.CloseSession (self.ck_cookie)
            self.ck_session = None

        if self.session_process is not None:
            self.session_process.wait()

        if self.guest_session:
            self.cleanup_guest_user()

        self.SessionClosed(self.ck_session_id)

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def Run( self, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.Run
        '''
        if self.session_worker is None:
            raise TaciXSessionException('SessionWorker not running')

        self.logger.debug('%r' % self.get_properties_complete())
        self.logger.debug('Sending Run to the Session Worker')
        self.session_worker.Run(self.ck_cookie, self.get_properties_complete(), reply_handler=self.__run_async_rec, error_handler=self.__async_error_handler)
        return True

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def Resume( self, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.Resume
        '''
        #self._check_polkit_privilege(sender, conn, 'org.tacix.session.Resume')
        if self.session_worker is None:
            raise TaciXSessionException('SessionWorker not running')
        self.logger.debug('Sending Run to the Session Worker')
        ret = self.session_worker.Resume(self.get_properties())
        self.logger.debug('Session Worker returned')
        return ret

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def Suspend( self, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.Suspend
        '''
        #self._check_polkit_privilege(sender, conn, 'org.tacix.session.Resume')
        if self.session_worker is None:
            raise TaciXSessionException('SessionWorker not running')
        ret = self.session_worker.Suspend()
        return ret

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='b', sender_keyword='sender', connection_keyword='conn' )
    def Terminate( self, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.Terminate
        '''
        #self._check_polkit_privilege(sender, conn, 'org.tacix.session.Resume')
        if self.session_worker is None:
            raise TaciXSessionException('SessionWorker not running')
        ret = self.session_worker.Terminate()
        return ret

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='a{ss}', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def AddPrinter( self, properties, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.AddPrinter \
          dict:string:string:"key0","value0","key1","value1"
        '''
        #self._check_polkit_privilege(sender, conn, 'org.tacix.session.Printer')
        self.logger.debug("Asking for printer with parameters %s" % properties)
        if 'printer' in properties:
            self.printers[properties['printer']] = properties
            self.PrinterSolicited(properties['printer'])

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='as', sender_keyword='sender', connection_keyword='conn' )
    def GetUnmanagedPrinters(self, sender, conn):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.GetUnmanagedPrinters
        '''
        ret = self.printers.keys()
        for printer_name in ret:
            if 'configured' in self.printers[printer_name]:
                ret.remove(printer_name)
        return ret

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='as', sender_keyword='sender', connection_keyword='conn' )
    def GetPrinters(self, sender, conn):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.GetPrinters
        '''
        ret = self.printers.keys()
        for printer_name in ret:
            if 'configured' not in self.printers[printer_name]:
                ret.remove(printer_name)
        return ret

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='a{ss}', out_signature='', sender_keyword='sender', connection_keyword='conn' )
    def AddMount( self, properties, sender, conn ):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.AddMount \
          dict:string:string:"key0","value0","key1","value1"
        '''
        #self._check_polkit_privilege(sender, conn, 'org.tacix.session.Mount')
        self.logger.debug("Asking for mount with parameters %s" % properties)
        if 'share' in properties:
            self.mounts[properties['share']] = properties
            self.MountSolicited(properties['share'])
        self.MountAdded(mount_name)

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='o', sender_keyword='sender', connection_keyword='conn' )
    def GetUnmanagedMounts(self, sender, conn):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.GetUnmanagedMounts
        '''
        ret = self.mounts.keys()
        for mount_name in ret:
            if self.mounts[mount_name].IsManaged():
                ret.remove(mount_name)
        return ret

    @dbus.service.method( dbus_interface=SESSION_INTERFACE,
                       in_signature='', out_signature='o', sender_keyword='sender', connection_keyword='conn' )
    def GetMounts(self, sender, conn):
        '''
        Example:
        dbus-send --system --dest=org.tacix \
          --type=method_call --print-reply --reply-timeout=2000 \
          /org/tacix/Manager \
          org.tacix.Manager.GetMounts
        '''
        ret = self.mounts.keys()
        return ret

    def _create_ck_session(self, uid, display, remote_host):
        try:

            # Defining the session attributes
            params = dbus.Array ([], signature = "(sv)")
            #params.append (("ssid", dbus.String (self.ck_session.GetId(), variant_level=1)))
            params.append (("unix-user", dbus.Int32 (uid, variant_level=1)))
            params.append (("session-type", dbus.String (SESSION_TYPE, variant_level=1)))
            params.append (("x11-display", dbus.String (display, variant_level=1)))
            params.append (("x11-display-device", dbus.String ("", variant_level=1)))
            params.append (("display-device", dbus.String ("", variant_level=1)))
            params.append (("remote-host-name", dbus.String (remote_host, variant_level=1)))
            params.append (("is-local", dbus.Boolean (True, variant_level=1)))

            # Create the ConsoleKit session
            self.ck_cookie = self.ck_manager.OpenSessionWithParameters (params)
            self.logger.debug (self.ck_cookie)

            # Getting the ConsoleKit session
            current_session = self.ck_manager.GetSessionForCookie (self.ck_cookie)

            session_obj = self.bus.get_object ('org.freedesktop.ConsoleKit', current_session)
            session = dbus.Interface (session_obj, 'org.freedesktop.ConsoleKit.Session')
            self.ck_session = session

            # Setting the session as active
            properties = dbus.Interface (session_obj, 'org.freedesktop.DBus.Properties')
            properties.Set ("org.freedesktop.DBus.Properties", "active", dbus.Boolean (True, variant_level=1))

            return self.ck_cookie

        except dbus.DBusException, e:
            # Dbus error problably you don't have the dbus rule installed or your launcher is not suid nx
            # Open session without the parameters
            self.logger.error ("Failed to create a CK session using parameters")

            error_string = str(e)
            self.logger.error ("Error: " + error_string)

            try:
                self.ck_cookie = self.ck_manager.OpenSession()
                self.logger.debug (self.ck_cookie)
                return self.ck_cookie
            except dbus.DBusException, e:
                self.logger.error ("Failed to create a CK session using parameters")

                error_string = str(e)
                self.logger.error ("Error: " + error_string)
                return ""


