#!/usr/bin/env python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os, os.path
import sys
import subprocess
import logging
import dbus

import dbus.mainloop.glib
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

import xdg.BaseDirectory

BUS_NAME="org.tacix"

MANAGER_INTERFACE="org.tacix.SessionManager"
MANAGER_PATH="/org/tacix/SessionManager"

SESSION_INTERFACE="org.tacix.Sessions"
SESSION_PATH="/org/tacix/Sessions/"
SESSION_TYPE="txsession"

def get_session(sessionid, bus=dbus.SystemBus()):
    manager_obj = bus.get_object (BUS_NAME, MANAGER_PATH)
    manager = dbus.Interface (manager_obj, MANAGER_INTERFACE)
    session_path = manager.GetSession (sessionid)
    session_obj = bus.get_object (BUS_NAME, session_path)
    return dbus.Interface (session_obj, SESSION_INTERFACE)

class TXPath:
    def __init__( self, sessionid=None, create=False, client=False ):
        if sessionid is None:
            if 'NXSESSIONID' not in os.environ:
                raise Exception("Session id not found")
            sessionid = os.environ['NXSESSIONID']
	self.session_id = sessionid

        if client:
            self.session_dir = os.path.join(os.environ['HOME'], '.nx/S-' + self.session_id)
        else:
            self.session_dir = os.path.join(os.environ['HOME'], '.nx/C-' + self.session_id)
        if os.path.exists(self.session_dir):
            return

        if not create:
            raise Exception( 'Session path %s not found' % self.session_dir )

        os.makedirs( self.session_dir, 0700 )
        last_dir = os.path.join(os.environ['HOME'], '.nx/C-last')
        if os.path.lexists(last_dir):
            os.remove(last_dir)
        os.symlink(self.session_dir, last_dir)
        for dir in ('pids', 'logs', 'services'):
            os.makedirs(os.path.join(self.session_dir, dir), 0700)

    def get_logger( self, progname ):
        log_filename = os.path.join( os.path.join(self.session_dir, "logs"), progname + '.log')
        log_filename2 = os.path.join( os.path.join(self.session_dir, "logs"), 'all.log')
        return get_logger( progname, log_filename, log_filename2  )

    def get_filepath( self, filename ):
        return os.path.join( self.session_dir, filename )

    def get_dir( self ):
        return self.session_dir

    def get_failed_dir( self ):
        return os.path.join(os.environ['HOME'], '.nx/F-' + self.session_id)

    def get_terminated_dir( self ):
        return os.path.join(os.environ['HOME'], '.nx/T-' + self.session_id)

    def get_unknown_dir( self ):
        return os.path.join(os.environ['HOME'], '.nx/U-' + self.session_id)

def opt_read(opt_str, sep=","):
    """Parse options into a dict."""

    options = {}
    if opt_str.find( sep ) is -1:
        return options
    for pair in opt_str.split( sep ):
        if pair.find( '=' ) is not -1:
            name, val = pair.split('=')
            options[name] = val
    return options

def get_tmp_logger( progname ):
    log_filename = os.path.join( "/tmp", progname + '.log')
    return get_logger( progname, log_filename )

def get_home_logger( progname ):
    log_filename = os.path.join( os.environ['HOME'], '.nx/' + progname + '.log')
    return get_logger( progname, log_filename )

def get_logger( progname, file, file2=None ):
    logger = logging.getLogger (progname)
    debug = False
    if xdg.BaseDirectory.load_first_config('tacix', "enable_debug") != None or os.path.exists("/etc/tacix/enable_debug"):
        debug = True
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

    if file is None:
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    else:
        handler = logging.FileHandler(file)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    if file2 is not None:
        handler2 = logging.FileHandler(file2)
        handler2.setFormatter(formatter)
        logger.addHandler(handler2)

    logger.debug ("Starting %s" % progname)
    return logger

def run_as_user(effective_pwd, session_id, session_type):
    if effective_pwd is None:
        return False
    pwd_user = effective_pwd

    new_env = {}
    new_env['USER'] = pwd_user.pw_name
    new_env['USERNAME'] = pwd_user.pw_name
    new_env['LOGNAME'] = pwd_user.pw_name
    new_env['HOME'] = pwd_user.pw_dir
    new_env['NXSESSIONID'] = session_id
    new_env['NX_SESSION_TYPE'] = session_type
    for env_var in ['LANG', 'LANGUAGE', 'PATH']:
        if env_var in os.environ:
            new_env[env_var] = os.environ[env_var]

    cmd = ["/bin/su", "-p", "-c", "/usr/lib/tacix/tacix-session", "-l", pwd_user.pw_name]
    return subprocess.Popen(cmd, env=new_env, close_fds=True, stdin=None, stdout=None, stderr=None)

